////////////////////////////////////////////////////////////////////////////////
//
//  Copyright (c) 2007 IWantGamez, LLC.  All rights reserved.
//  
//  Author: John Holben         Date: 04.28.2007
//
////////////////////////////////////////////////////////////////////////////////

// Constants 
var MAX_GRID = 5;
var GEM_SIZE_Y = 23;
var GEM_SIZE_X = 23;
var MAX_GEM = 4;
var OFFSET_X = 7;
var OFFSET_Y = 21;
var SELECTOR_IMAGE = "Graphics/selector.png";
var MATCHED_IMAGE = "Graphics/matched.png";
var SELECTOR_DEFAULT = 2;
var LOOP_INTERVAL = 100; // 10 fps
var GS_IDLE = 0;
var GS_SWAP = 1;
var GS_END_SWAP = 2;
var GS_MATCH = 3;
var GS_END_MATCH = 4;
var GS_FINISHED = 5;
var GS_DROP = 6;
var GS_END_DROP = 7;
var GS_GAME_OVER = 8;
var GS_END_GAME_OVER = 9;

// Audio
var WAV_SWAP = 0;
var WAV_MATCH = 1;
var WAV_HINT = 2;
var WAV_GAME_OVER = 3;
var WAV_GAME_START = 4;
var WAV_GAME_QUIT = 5;
var WAV_HI_SCORE = 6;
var WAV_BONUS = 7;

// From settings.js
var KEY_ENABLE_SOUND_FX = "SoundFX";
var KEY_HS_TIME = "HSTime";
var KEY_HS_GEMS = "HSGems";
var KEY_HS_MOVES = "HSMoves";
var KEY_HS_HINTS = "HSHints";
var KEY_EXISTS = "EXISTS";

// Globals
var IMAGE = new Array(MAX_GEM);
var LEVEL = new ArrayTemplate(MAX_GRID,MAX_GRID);
var MATCHED = new ArrayTemplate(MAX_GRID,MAX_GRID);
var ANIMATION = new Animation();
var GEMS = new Array();
var STATS = new StatTemplate();
var AUDIO = new AudioTemplate();
var SELECTOR; 
var TOTAL_SECONDS = 0;
var GAME_LOOP;
var GAME_OVER = false;

// Debug
var DEBUG = 0;

IMAGE[0] = "Graphics/ruby.png";
IMAGE[1] = "Graphics/amathist.png";
IMAGE[2] = "Graphics/emerald.png";
IMAGE[3] = "Graphics/yellow.png";

function Initialize()
{
    // Get audio going 
    AUDIO.Init();
    // Monitor incase player turns off soundFx
    System.Gadget.onSettingsClosed = AUDIO.SettingsClosed;
    // Create the Game Loop 
    GAME_LOOP = new GameLoopTemplate();
    // Create Selector and default to middle of game board
    SELECTOR = new SelectorTemplate();
    // Create random level
    ConsolidateLevel();
    // Start the Game Loop running 
    setTimeout(GAME_LOOP.Run,1);
    // This is required so control is focused and Tool tips will display
    // before any clicking of the game pieces
    self.focus();
    document.body.focus();
}

function IsUnmatchedGem(r,c,t)
{
    // Check above
    var gem = LEVEL.Get(r -1,c);    
    if(gem != null)
       if(gem.Type == t)
            return false;    

    // Check below
    gem = LEVEL.Get(r + 1,c);    
    if(gem != null)
       if(gem.Type == t)
            return false;
    
    // Check left
    gem = LEVEL.Get(r,c - 1);    
    if(gem != null)
       if(gem.Type == t)
           return false;
    
    // Check right
    gem = LEVEL.Get(r,c + 1);    
    if(gem != null)
       if(gem.Type == t)
           return false;    
        
    // No Matching Gems
    return true;
}

function AddGem(g,m)
{
    // Make Sure Gem isn't already in list
    for(var i = 0; i < m.length; i++)
    {
        if(m[i] == g)
        {
            // Already in the list
            return;
        }
    }
    // Ok to add
    m.push(g);
}

function ProcessMatches()
{
    var gem;
    var foundMatch = false;
   // Check each Gem for matches
    for(var r=0;r < MAX_GRID;r++)
        for(var c=0;c < MAX_GRID;c++)
        {
            gem = LEVEL.Get(r,c);
            if(gem != null && gem.State == GS_IDLE)
            {
                var location;
                var temp;
                var verticalMatches = new Array();   
                var horizontalMatches = new Array();
                
                // Add selected gem to lists
                verticalMatches.push(gem);
                horizontalMatches.push(gem);
                
                CheckForMatch(gem,horizontalMatches,verticalMatches);
                
                // Did we find any horizontal matches?
                if(horizontalMatches.length > 2)
                {
                    foundMatch = true;
                    for(var i = 0; i < horizontalMatches.length; i++)
                    {
                        location = horizontalMatches[i].Location;
                        temp = MATCHED.Get(location.Row,location.Col);
                        if(temp == null)            
                            MATCHED.Set(location.Row,location.Col,new MatchedTemplate(location.Row,location.Col));
                        
                        // Initialize the Match Animation - Set State to GS_MATCH
                        horizontalMatches[i].Match();
                    }
                }
                
                // Did we find any vertical matches?
                if(verticalMatches.length > 2)
                {
                    foundMatch = true;
                    for(var i = 0; i < verticalMatches.length; i++)
                    {
                        location = verticalMatches[i].Location;
                        temp = MATCHED.Get(location.Row,location.Col);
                        if(temp == null)            
                            MATCHED.Set(location.Row,location.Col,new MatchedTemplate(location.Row,location.Col));
                        
                        // Initialize the Match Animation - Set State to GS_MATCH
                        verticalMatches[i].Match();
                    }
                }
            }
        }

    return foundMatch;
}

function CheckForMatch(checkGem,horizontalMatches,verticalMatches)
{
    var type = checkGem.Type;
    
    // Check above
    var r = checkGem.Location.Row - 1;
    var c = checkGem.Location.Col;
    var gem = LEVEL.Get(r,c);
    
    while(gem !=null && type == gem.Type && gem.State == GS_IDLE)
    {
        AddGem(gem,verticalMatches);                                   
        gem = LEVEL.Get(--r,c);
    }
    
    // Check below
    r = checkGem.Location.Row + 1;
    c = checkGem.Location.Col;
    gem = LEVEL.Get(r,c);
    
    while(gem !=null && type == gem.Type && gem.State == GS_IDLE)
    {
        AddGem(gem,verticalMatches);                                   
        gem = LEVEL.Get(++r,c);
    }

    // Check left
    r = checkGem.Location.Row;
    c = checkGem.Location.Col - 1;
    gem = LEVEL.Get(r,c);
    
    while(gem !=null && type == gem.Type && gem.State == GS_IDLE)
    {
        AddGem(gem,horizontalMatches);                                   
        gem = LEVEL.Get(r,--c);
    }

    // Check right
    r = checkGem.Location.Row;
    c = checkGem.Location.Col + 1;
    gem = LEVEL.Get(r,c);
    
    while(gem !=null && type == gem.Type && gem.State == GS_IDLE)
    {
        AddGem(gem,horizontalMatches);                                   
        gem = LEVEL.Get(r,++c);
    }
}

function ConsolidateLevel()
{
    var madeChange = false;
    for(r=MAX_GRID - 1;r >= 0; r--)
        for(c=0;c < MAX_GRID;c++)
        {
            var gem = LEVEL.Get(r,c);

            if(gem == null)
            {
                madeChange = true;
                if(r == 0)
                {
                    var newGem = 0;
                    for(t =0; t< 100; t++)
                    {
                        newGem = SELECTOR.GetNextType();
                        // Make sure inital level has no matches
                        if(IsUnmatchedGem(0,c,newGem) == true)
                            break;
                    }
                    gem = new GemTemplate(r,c,newGem);
                        
                    // Make sure we are positioned correctly                    
                    gem.Element.style.top = "0px";
                    gem.Element.style.left = ((c * GEM_SIZE_X) + OFFSET_X) + "px";

                    // This will init us and add us to the Level
                    gem.Drop(r,c);
                }
                else
                {
                    gem = LEVEL.Get(r-1,c);
                    if(gem != null)
                        gem.Drop(r,c);                                    
                }            
            }
        }

    return madeChange;
}

function hint()
{
    AUDIO.Play(WAV_HINT);
    STATS.Hints++;
    
    var foundMatch = false;
    var current = null;
    for(var r = 0; r < MAX_GRID && foundMatch == false; r++)
        for(var c =0; c < MAX_GRID; c++)
        {
             current = LEVEL.Get(r,c);
            
            // look Right 
            
            //   X
            //  XOX 
            //   X
            if((gem = LEVEL.Get(r-1,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r,c+2)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r,c+1);  break; }

            if((gem = LEVEL.Get(r+1,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r,c+2)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r,c+1);  break; }
                            
            //    X
            //  XXO 
            //    X        
            if((gem = LEVEL.Get(r,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r-1,c+2)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r,c+2);  break; }
        
            if((gem = LEVEL.Get(r,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+1,c+2)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r,c+2);  break; }

            //   XX
            //  XOO 
            //   XX        
            if((gem = LEVEL.Get(r-1,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r-1,c+2)) != null)
                        if(gem.Type == current.Type) { foundMatch = true;  break; }

            if((gem = LEVEL.Get(r+1,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+1,c+2)) != null)
                        if(gem.Type == current.Type) { foundMatch = true;  break; }

            
            //  XXOX 
            if((gem = LEVEL.Get(r,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r,c+3)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r,c+2);  break; }

            //  XOXX 
            if((gem = LEVEL.Get(r,c+2))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r,c+3)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r,c+1);  break; }
        
        
            // look Down 
            
            //   X
            //  XOX 
            //   X
            if((gem = LEVEL.Get(r+1,c-1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+2,c)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r+1,c);  break; }
        
            if((gem = LEVEL.Get(r+1,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+2,c)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r+1,c);  break; }
        
        
            //   X
            //   X 
            //  XOX
            if((gem = LEVEL.Get(r+1,c))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+2,c-1)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r+2,c);  break; }
        
            if((gem = LEVEL.Get(r+1,c))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+2,c+1)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r+2,c);  break; }
        

            //   X
            //  XOX 
            //  XOX
            if((gem = LEVEL.Get(r+1,c-1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+2,c-1)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; break; }
        
            if((gem = LEVEL.Get(r+1,c+1))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+2,c+1)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; break; }


            //  X
            //  X 
            //  O
            //  X
            if((gem = LEVEL.Get(r+1,c))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+3,c)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r+2,c); break; }


            //  X
            //  O 
            //  X
            //  X
            if((gem = LEVEL.Get(r+2,c))!=null)
                if(gem.Type == current.Type)
                    if((gem = LEVEL.Get(r+3,c)) != null)
                        if(gem.Type == current.Type) { foundMatch = true; current = LEVEL.Get(r+1,c); break; }
        }

    
    if(foundMatch == true)
        SELECTOR.Select(current.Location);

}

function GameOver()
{
    hintButton.style.visibility= "hidden";
    // You Won!! 
    GAME_OVER = true;
    // Trigger domino effect
    LEVEL.Get(0,0).GameOver();
}

function CreateGameOverScreen()
{
    var setHiScore = false;
    var hiScoreTime = System.Gadget.Settings.readString(KEY_HS_TIME);
    // Is there a high score saved?
    if(hiScoreTime.length < 8)
    {
        setHiScore = true;
    }
    else
    {
        // convert time to seconds 
        var seconds = parseInt(hiScoreTime.substr(0,2)) * 60 * 60;
        seconds += (parseInt(hiScoreTime.substr(3,2)) * 60);
        seconds += parseInt(hiScoreTime.substr(6,2));
        // Did we beat the record?
        if(TOTAL_SECONDS < seconds)
            setHiScore = true;
    }

    if(setHiScore == true)
    {
        // Write out new high score!
        System.Gadget.Settings.writeString(KEY_HS_TIME,timerCount.innerText);
        System.Gadget.Settings.writeString(KEY_HS_GEMS,STATS.Gems);
        System.Gadget.Settings.writeString(KEY_HS_MOVES,STATS.Moves);
        System.Gadget.Settings.writeString(KEY_HS_HINTS,STATS.Hints);
        System.Gadget.Settings.write(KEY_EXISTS, true);
        // Show special bonus screen
        trophie.style.visibility = "visible";
        AUDIO.Play(WAV_HI_SCORE);
    }
    else
    {
        AUDIO.Play(WAV_GAME_OVER);
        wonMessage.style.visibility = "visible";
    }

    txtTotalTime.innerText = timerCount.innerText;
    txtGems.innerText = STATS.Gems;
    txtMoves.innerText = STATS.Moves;
    txtHints.innerText = STATS.Hints;
    
    GAME_LOOP = new GameOverLoopTemplate();
    setTimeout(GAME_LOOP.Run, LOOP_INTERVAL);
}

function GameOverLoopTemplate()
{
    this.State = 0;
    this.Count = 0;
    
    this.Run = function() 
        {
            switch(this.State)
            {
                case 0:
                {
                    var finished = true;
                    for(var r = 0; r < MAX_GRID; r++)
                        for(var c = 0; c < MAX_GRID; c++)
                        {
                            var g = MATCHED.Get(r,c);
                            var width = parseInt(g.Element.style.width) - 1;
                            var left = parseInt(g.Element.style.left);
                            
                            if(width >= 0)
                            {
                                finished = false;
                                
                                if(r % 2)
                                {
                                    if(c % 2)
                                    {
                                        left++;
                                        g.Element.style.left = left + "px";
                                    }

                                    g.Element.style.width = width + "px";
                                }
                                else
                                {
                                    if((c % 2) == 0)
                                    {
                                        left++;
                                        g.Element.style.left = left + "px";
                                    }
                                    g.Element.style.width = width + "px";
                                }
                            }
                        }
                        
                        if(finished == true)
                        {
                            this.State = 1;
                            this.Count = 0;
                        }
                        
                        break;
                }
                case 1:
                {
                    if(++this.Count > 15)
                    {   
                        this.Count = 0;
                        
                        // We are finished
                        if(trophie.style.visibility == "visible")
                        {
                            gameOver.filters[0].Apply();
                            trophie.style.visibility = "hidden";
                            wonMessage.style.visibility = "visible";
                            gameOver.filters[0].Play();
                            AUDIO.Play(WAV_GAME_OVER);
                        }
                        return;   
                    }
                    break;    
                }
            }        
            // Restart Loop
            setTimeout(function() {GAME_LOOP.Run();}, LOOP_INTERVAL);
      }
}


function GameLoopTemplate()
{
    this.lastDate = new Date();
    this.totalTicks = 0;
    this.subTick = 0;
    this.tempGems = 0;

    this.Run = function() 
        {
            if(DEBUG == 1) { debugger; DEBUG = 0; }
            
            var now = new Date();
            var ticks = now - GAME_LOOP.lastDate;
            this.lastDate = now;
            GAME_LOOP.totalTicks += ticks;
            GAME_LOOP.subTick += ticks;
            
            if(GAME_LOOP.subTick > 1000)
            {
                var seconds = parseInt(GAME_LOOP.subTick / 1000);
                GAME_LOOP.subTick %= 1000;
                
                if(GAME_OVER == false)
                    TOTAL_SECONDS += seconds;    
            }
            
            // Update Status Display
            var timeCount = Math.abs(TOTAL_SECONDS);
            var hours = parseInt(timeCount / 3600);
            var minutes = parseInt((timeCount / 60) % 60);
            var seconds = parseInt(timeCount % 60);
            timerCount.innerText = (hours < 10 ? "0" : "")+hours+":"+(minutes < 10 ? "0" : "")+minutes+":"+(seconds < 10 ? "0" : "")+seconds;
            
            if(this.tempGems < STATS.Gems)
                this.tempGems++;
            
            var gems = "0000" + this.tempGems;
            gemCount.innerText = gems.substr(gems.length - 4,4);
            
            var timerAccuracy = ticks / LOOP_INTERVAL;
            
            // Process Selector 
            SELECTOR.Tick(timerAccuracy);
            // Process Game Pieces
            LEVEL.Tick(timerAccuracy);

            if(LEVEL.CheckForMatches == true)
            {
                LEVEL.CheckForMatches = false;

                if(ProcessMatches() == false)
                    SELECTOR.UnSwap();
                else
                    AUDIO.Play(WAV_MATCH);
            }


            // This will only be set if Gem has FINISHED removing itself            
            if(LEVEL.Consolidate == true)
            {
                LEVEL.Consolidate = false;
                
                if(GAME_OVER == false)
                {
                    if(ConsolidateLevel() == false)
                        if(ProcessMatches() == false)
                        {
                            if(MATCHED.IsFull() == true)
                                GameOver();
                        }
                        else
                        {
                            AUDIO.Play(WAV_MATCH);
                        }
                }
                else
                {
                    if(LEVEL.IsEmpty() == true)
                    {
                        // At this point all Jems are off the screen                    
                        CreateGameOverScreen();
                        // end the loop!
                        return;
                    }                            
                }
            }
            
            // Restart Loop
            setTimeout(function() {GAME_LOOP.Run();}, LOOP_INTERVAL);
        }
}

function InitGemSwap(r,c)
{
    AUDIO.Play(WAV_SWAP);
    
    this.State = GS_SWAP;
    
    if(r == this.Location.Row)
    {
        if(c > this.Location.Col)
            this.InitAnimation(ANIMATION.horzRight);            
        else
            this.InitAnimation(ANIMATION.horzLeft);            
    }
    else if(r > this.Location.Row)
        this.InitAnimation(ANIMATION.vertDown);    
    else
        this.InitAnimation(ANIMATION.vertUp);    
    
    // Set up animation
    this.Location.Row = r;
    this.Location.Col = c;
}

function InitGemDrop(r,c)
{
    this.State = GS_DROP;
    // Remove us from current location
    LEVEL.Set(this.Location.Row,this.Location.Col,null);
    // move us to new position in Level
    LEVEL.Set(r,c,this);
    // Update our location
    this.Location.Row = r;
    this.Location.Col = c;
    // Animation will move us to the destination
    this.InitAnimation(ANIMATION.Drop);
}

function InitGemMatch()
{
    this.State = GS_MATCH;
    // Remove shadow - removeObjects doesn't seem to work! 
    this.Element.addShadow("black",0,100,0,0);
    this.Element.brightness = 180;
    // Set up animation
    this.InitAnimation(ANIMATION.Match);
    this.Element.style.filter = "progid:DXImageTransform.Microsoft.Pixelate(MaxSquare=8,duration=1,enabled=true)"; 
    this.Element.filters[0].apply();
    this.Element.filters[0].play();
    // Update stat
    STATS.GemRemoved();
}

function InitGemGameOver()
{
    if(this.Location.Row == this.Location.Col)
        AUDIO.Play(WAV_BONUS);
    this.State = GS_GAME_OVER;
    // Set up animation
    // Remove shadow - removeObjects doesn't seem to work! 
    this.Element.addShadow("black",0,100,0,0);
    this.Element.brightness = 180;
    // Set up animation
    this.InitAnimation(ANIMATION.GameOverStart);
    this.Element.style.filter = "progid:DXImageTransform.Microsoft.Pixelate(MaxSquare=8,duration=2,enabled=true)"; 
    this.Element.filters[0].apply();
    this.Element.filters[0].play();

}

function GemTick(timerAccuracy)
{
    switch(this.State)
    {
        case GS_IDLE:
        case GS_FINISHED:
            // nothing to do
            break;
        case GS_SWAP:

            // run animation
            if(this.RunAnimation() == false)
                this.State = GS_END_SWAP;
    
            break;
            
        case GS_END_SWAP:
            
            this.Element.style.left = this.Location.GetX();
            this.Element.style.top = this.Location.GetY();
            this.State = GS_IDLE;                    
            LEVEL.CheckForMatches = true;        
            break;

        case GS_MATCH:
            
            if(this.RunAnimation() == false)
                this.State = GS_END_MATCH;
    
            break;
        
        case GS_END_MATCH:
        
            // Level will remove us with this state
            this.State = GS_FINISHED;
            // make sure filter has stopped
            this.Element.filters(0).stop();
            // Remove us from the page
            document.body.removeChild(this.Element);
            // Remove us from the level
            LEVEL.Set(this.Location.Row,this.Location.Col,null);
            // Notify level it should consolidate
            LEVEL.Consolidate = true;
            break;
    
        case GS_DROP:
            
            if(this.RunAnimation() == false)
                this.State = GS_END_DROP;
            
            break;
            
        case GS_END_DROP:
        
            // Make sure we are located properly
            this.Element.style.left = this.Location.GetX();
            this.Element.style.top = this.Location.GetY();
            this.State = GS_IDLE;
            LEVEL.Consolidate = true;
            break;
        case GS_GAME_OVER:
                    
            if(this.RunAnimation() == false)
                if(this.Frames == ANIMATION.GameOverFinish)
                {
                    this.State = GS_END_GAME_OVER;
                }
                else
                {
                    // Start Cascade...
                    var gem = LEVEL.Get(this.Location.Row,this.Location.Col + 1);
                    if(gem != null)
                        gem.GameOver();
                    
                    gem = LEVEL.Get(this.Location.Row + 1,this.Location.Col);
                    if(gem != null)
                        gem.GameOver();
                
                    // Finish our animation
                    this.InitAnimation(ANIMATION.GameOverFinish);                
                }
            break;
            
        case GS_END_GAME_OVER:

            // Use same notification to signal game loop we are finished
            this.State = GS_END_MATCH;
            break;
    }
}

function GemTemplate(r,c,t)
{
    // Public Properties
    this.Location = new LocationTemplate(r,c);
    this.Type = t;
    this.State = 0;
    this.State = GS_IDLE;
    this.dragX = 0;
    this.dragY = 0;
    
    // Animation
    this.Frames = null;
    this.StartTick = 0;
    this.InitAnimation = function(frames) { this.Frames = frames; this.StartTick = GAME_LOOP.totalTicks; this.Location.X = parseInt(this.Element.style.left); this.Location.Y = parseInt(this.Element.style.top); };
    this.RunAnimation = function() { return ANIMATION.ApplyFrame(GAME_LOOP.totalTicks - this.StartTick,this.Frames,this);  };
    
    
    // Create Element and Initialize
    this.Element = document.createElement("<g:image id='gem" + ((this.Location.Row * MAX_GRID) + this.Location.Col) + "' dir='ltr' style='position:absolute; z-index:15;left:" + this.Location.GetX() + ";top:" + this.Location.GetY() + ";' ></g:image>"); // width:" + GEM_SIZE_X + "pt;height:" + GEM_SIZE_Y + "pt;
    this.Element.src = IMAGE[t];
    //this.Element.style.width = GEM_SIZE_X;
    //this.Element.style.height = GEM_SIZE_Y;;
    this.Element.addShadow("black",4,50,2,2);
    this.Element.opacity = 75;
    // Setup back pointer for onclick
    this.Element.Gem = this;
    // Set up simple mouse over highlight
    this.Element.onmouseover = function() { this.addGlow('white',2,40); }; 
    this.Element.onmouseout = function() { this.addGlow('white',0,0); };
    this.Element.onclick = function() { SELECTOR.Select(this.Gem.Location); };
    
    // Add us to page 
    document.body.appendChild(this.Element);

    // Shared Methods
    this.Tick = GemTick;
    this.Swap = InitGemSwap;
    this.Drop = InitGemDrop;
    this.Match = InitGemMatch;
    this.GameOver = InitGemGameOver;
}

function MatchedTemplate(r,c)
{
    this.Location = new LocationTemplate(r,c);
    
    this.Element = document.createElement("<g:image id='gem" + ((this.Location.Row * MAX_GRID) + this.Location.Col) + "' style='position:absolute; z-index:10;left:" + this.Location.GetX() + ";top:" + this.Location.GetY() + ";' ></g:image>"); 
    this.Element.src = MATCHED_IMAGE;
    document.body.appendChild(this.Element);
}

function SelectorTemplate()
{
    this.Rotation = 0;
    this.Location = new LocationTemplate(-1,-1);
    this.Element = document.createElement("<g:image id='selector' style='position:absolute; z-index:14;left:" + this.Location.GetX() + ";top:" + this.Location.GetY() + ";visibility:hidden;' ></g:image>"); 
    this.Element.src = SELECTOR_IMAGE;
    document.body.appendChild(this.Element);
    this.SwapGem1 = null;
    this.SwapGem2 = null;
    this.Visible = false;
    this.LastType = 0;
    
    
    this.Select = function(l) 
                    { 
                        if(GAME_OVER == true)
                            return;
                    
                        // Calculate the selected Row Col
                        var row = l.Row; 
                        var col = l.Col; 
                        var wasVisible = this.Visible;
                        
                        // If visible is false then this is first selection    
                        if(this.Visible == false)
                        {
                            this.Element.style.visibility = "";
                            this.Visible = true;
                        }

                        // Should we try to swap or should we reposition the selector?
                        if( (((this.Location.Row == (row -1) || this.Location.Row == (row + 1)) && this.Location.Col == col ) || 
                            ((this.Location.Col == (col - 1) || this.Location.Col == ( col + 1)) && this.Location.Row == row )) && wasVisible == true )
                            {
                                STATS.Moves++;
                            
                                // swap gems
                                this.SwapGem1 = LEVEL.Get(row,col);
                                this.SwapGem2 = LEVEL.Get(this.Location.Row,this.Location.Col);
                                
                                // Update the LEVEL Array
                                LEVEL.Set(row,col,this.SwapGem2);
                                LEVEL.Set(this.Location.Row,this.Location.Col,this.SwapGem1);
                                
                                // Have each Gem animate
                                this.SwapGem1.Swap(this.Location.Row,this.Location.Col)
                                this.SwapGem2.Swap(row,col);
                                
                                // We are finished so hide
                                this.Element.style.visibility = "hidden";
                                this.Visible = false;
                                
                            }
                            else
                            {
                                // to much
                                // AUDIO.Play(WAV_CLICK);
                                this.Location.Row = row;  
                                this.Location.Col = col; 
                                this.Element.style.left = this.Location.GetX();
                                this.Element.style.top  = this.Location.GetY();
                            }
                            
                    };
    
    this.UnSwap =   function()
                    {
                        // un-swap gems
                        
                        if(this.SwapGem1 == null)
                            return;
                            
                        // To much
                        // AUDIO.Play(WAV_UN_SWAP);
                        
                        // Update the LEVEL Array
                        LEVEL.Set(this.SwapGem1.Location.Row,this.SwapGem1.Location.Col,this.SwapGem2);
                        LEVEL.Set(this.SwapGem2.Location.Row,this.SwapGem2.Location.Col,this.SwapGem1);
                        
                        // Have each Gem animate
                        var gem1Row = this.SwapGem1.Location.Row;
                        var gem1Col = this.SwapGem1.Location.Col;
                        this.SwapGem1.Swap(this.SwapGem2.Location.Row,this.SwapGem2.Location.Col)
                        this.SwapGem2.Swap(gem1Row,gem1Col);
                        
                        this.SwapGem1 = null;
                        this.SwapGem2 = null;
                    };
    
    
    this.Tick =     function(timerAccuracy) 
                    {    
                        // Don't waste CPU if not visible
                        if(this.Visible == true)
                        {
                            this.Rotation += (10 * timerAccuracy);
                            if(this.Rotation > 360)
                                 this.Rotation = this.Rotation - 360;
                            this.Element.rotation = this.Rotation;
                        }
                    };


    this.GetNextType = function() {var temp = 0;  do{ temp = Math.round(Math.random() * (MAX_GEM - 1)); }while(this.LastType == temp); return this.LastType = temp; }    
}

function LocationTemplate(r,c)
{
    this.Row = r;
    this.Col = c;
    // Used only with animation for temp storage
    this.X = 0;
    this.Y = 0;
    
    this.GetX = function() { return (this.Col * GEM_SIZE_X) + OFFSET_X; } 
    this.GetY = function() { return (this.Row * GEM_SIZE_Y) + OFFSET_Y; }
}

function ArrayTemplate(maxRow,maxCol)
{
    this.Row = maxRow;
    this.Col = maxCol;
    this.Consolidate = false;
    this.CheckForMatches = false;
    
    this.array = new Array(maxRow * maxCol);
    
    this.Tick = function(ta) { for( var i =0; i < this.array.length; i++) { if(this.array[i] != null){this.array[i].Tick(ta);} } };
    this.Get = function(r,c) { if(r < 0 || r >= this.Row)return null; if(c < 0 || c >= this.Col)return null; return this.array[ (r * this.Row) + c ]; };
    this.Set = function(r,c,o) { this.array[ (r * this.Row) + c ] = o; }; 
    this.IsFull = function() { for(var i =0; i < this.array.length; i++) { if(this.array[i] == null) return false; } return true; };
    this.IsEmpty = function() { for(var i =0; i < this.array.length; i++) { if(this.array[i] != null) return false; } return true; };

}

function StatTemplate()
{
    this.Hints = 0;
    this.Gems = 0;
    this.Moves = 0;
    
    this.GemRemoved = function() { if(++this.Gems == 10000){ this.Gems =0; GAME_LOOP.tempGems = 0;} };
    
    this.IsHighScore = function() 
        { 
            // Is total time less then recorded total time? 
            return true;
        };
}

function AudioTemplate()
{
    this.UseSoundFx = true;

    this.Waves = new Array( "swap.wav", "match.wav", "hint.wav", "gameOver.wav", "gameStart.wav", "gameQuit.wav", "hiScore.wav", "bonus.wav");
    
    this.Init = function()
        {
            if(System.Gadget.Settings.read(KEY_EXISTS) == true)
            {
                this.UseSoundFx = System.Gadget.Settings.read(KEY_ENABLE_SOUND_FX);        
            }
        }
    
    this.Play = function(id)
        {
            if(this.UseSoundFx == true && id >= 0 && id < this.Waves.length)
            {
                System.Sound.playSound("Waves\\" + this.Waves[id]);            
            }
        }
        
        
    this.SettingsClosed = function(event)
        {
        
            if (event.closeAction == event.Action.commit)
            {
                AUDIO.UseSoundFx = System.Gadget.Settings.read(KEY_ENABLE_SOUND_FX); 
                
            }
        }
}


function Animation( )
{
    this.horzRight = [  
                     { "Key":    0, "X":  0, "Y":  0 } , 
                     { "Key":  150, "X": 11, "Y":  2 } ,
                     { "Key":  250, "X": 21, "Y":  0 } 
                  ];

    this.horzLeft = [
                     { "Key":    0, "X":   0, "Y":  0 } , 
                     { "Key":  150, "X": -11, "Y": -2 } ,
                     { "Key":  250, "X": -21, "Y":  0 } 
                  ];

    this.vertUp = [  
                     { "Key":    0, "X":   0, "Y":   0 } , 
                     { "Key":  150, "X":  -2, "Y": -10 } ,
                     { "Key":  250, "X":   0, "Y": -21 } 
                  ];

    this.vertDown = [
                     { "Key":    0, "X":  0, "Y":  0 } , 
                     { "Key":  150, "X":  2, "Y": 10 } ,
                     { "Key":  250, "X":  0, "Y": 21 } 
                  ];

    this.Drop = [
                     { "Key":    0, "X":  0, "Y":  0 } , 
                     { "Key":  200, "X":  0, "Y": 15 } ,
                     { "Key":  300, "X":  0, "Y": 20 } 
                ];

    this.Match = [
                     { "Key":    0, "X":  0, "Y":  0 } , 
                     { "Key":  360, "X":  0, "Y": -4 } 
                ];

    this.GameOverStart = [
                     { "Key":    0, "X":  0, "Y":  0 } , 
                     { "Key":  220, "X": -2, "Y": -2 } 
                ];

    this.GameOverFinish = [
                     { "Key":    0, "X": -4, "Y": -4 } , 
                     { "Key":  220, "X": -6, "Y": -6 } 
                ];

    this.ApplyFrame = function(tick,frames,gem) 
                {
                    var key1 = 0;
                    var key2 = 0;
                    
                    // Find what frame we are at
                    for( var i =0; i < frames.length; i++)
                    {
                        if(tick >= frames[i].Key)
                            key1 = i;
                    }
                    
                    // Define second frame
                    key2 = (key1 >= ( frames.length - 1)) ? key1 : key1 + 1; 
                    
                    // Get difference in key times
                    var timeDiff = frames[key2].Key - frames[key1].Key;
                    if(timeDiff < 1) 
                        timeDiff = 1;
                    
                    // calculate scalar
                    var scalar =  (tick - frames[key1].Key) / timeDiff;
                    
                    // calculate X with scaled value
                    var x = (frames[key2].X - frames[key1].X);
                    x *= scalar;
                    x += frames[key1].X;

                    // calculate Y with scaled value
                    var y = (frames[key2].Y - frames[key1].Y);
                    y *= scalar;
                    y += frames[key1].Y;
                
                    // apply to Element 
                    gem.Element.style.left = Math.round(gem.Location.X + x);
                    gem.Element.style.top = Math.round(gem.Location.Y + y);

                    // Are we finished?
                    return (key1 != key2); // (tick < frames[frames.length -1].Key);
                }
}


// For Debugging only 
function KeyHandler()
{
    // Press F5 to attach debugger 
    if (event.keyCode == 116)
    {
        DEBUG = 1;
    }


}